"""Command line script to clip a network"""

import argparse
import textwrap

from reticuler.system import System
from reticuler.backward_evolution.system_back import BackwardSystem
from reticuler.user_interface import clippers

# %%
def main():
    parser = argparse.ArgumentParser(
        description="Clip a network.", formatter_class=argparse.RawTextHelpFormatter
    )

    # defining arguments for parser object
    parser.add_argument(
        "input_file",
        type=str,
        nargs=1,
        metavar="file_name",
        help=textwrap.dedent(
            """\
                            File to import"""
        ),
    )       
    parser.add_argument(
        "-out",
        "--output_file",
        type=str,
        nargs=1,
        metavar="file_name",
        help=textwrap.dedent(
            """\
                            File to export. If None the same as input (+"_clipped" at the end).
                            default = None """
        ),
        default=None,
    )
        
    parser.add_argument(
        "-in_BEA",
        "--input_file_BEA",
        type=str,
        nargs=1,
        metavar="file_name",
        help=textwrap.dedent(
            """\
                            File with the BEA results to import"""
        ),
    ) 
    parser.add_argument(
        "-out_BEA",
        "--output_file_BEA",
        type=str,
        nargs=1,
        metavar="file_name",
        help=textwrap.dedent(
            """\
                            The BEA file to export. If None the same as input (+"_clipped" at the end).
                            default = None """
        ),
        default=None,
    )
    # Clipping type and limit
    parser.add_argument(
        "-S",
        "--step",
        type=float,
        nargs=1,
        metavar="num",
        help=textwrap.dedent(
            """\
                            Maximum step in the network evolution. If filled clips to step.
                            default = None
                            """
        ),
        default=None,
    )
    parser.add_argument(
        "-L",
        "--length",
        type=float,
        nargs=1,
        metavar="num",
        help=textwrap.dedent(
            """\
                            Maximum length of the network. If filled clips to length.
                            default = None
                            """
        ),
        default=None,
    )
    parser.add_argument(
        "-H",
        "--height",
        type=float,
        nargs=1,
        metavar="num",
        help=textwrap.dedent(
            """\
                            Maximum height of the network. If filled clips to height.
                            default = None
                            """
        ),
        default=None,
    )
    parser.add_argument(
        "-BEA",
        "--BEA_step",
        type=float,
        nargs=1,
        metavar="num",
        help=textwrap.dedent(
            """\
                            Clipping Maximum height of the network. If filled clips to height.
                            default = None
                            """
        ),
        default=None,
    )        

    # parse the arguments from standard input
    args = parser.parse_args()

    # Import System from JSON file
    system = System.import_json(input_file=args.input_file[0])
    
    if args.output_file is None:
        system.exp_name = args.input_file[0] + "_clipped"
    else:
        system.exp_name = args.output_file[0]
    
    if args.step is not None:
        clippers.clip_to_step(system, args.step[0])
    elif args.length is not None:
        clippers.clip_to_length(system, args.length[0])
    elif args.height is not None:
        clippers.clip_to_height(system, args.height[0])
    elif args.BEA_step is not None:
        backward_system = BackwardSystem.import_json(input_file=args.input_file_BEA[0], system=system)
        clippers.clip_to_BEA_step(system, backward_system, max_BEA_step=args.BEA_step[0])
        
        if args.output_file_BEA is None:
            backward_system.exp_name = args.input_file_BEA[0]+"_clipped"
        else:
            backward_system.exp_name = args.output_file_BEA[0]
              
        backward_system.export_json()          
    else:
        print("Network not clipped - you must choose one clipping limit!")

    system.export_json()


if __name__ == "__main__":
    main()



"""Solvers for PDEs describing the field around the network.

Classes:
    FreeFEM
    
"""

import numpy as np
import subprocess
import os.path
import os
from platform import system
from tempfile import NamedTemporaryFile
import textwrap


class FreeFEM:
    """PDE solver based on the finite element method implemented in FreeFEM [Ref2]_.

    Attributes
    ----------
    equation : int, default 0
        - 0: Laplace
        - 1: Poisson
    a1a2a3_coefficients : array
        An array of a1a2a3 coefficients for each tip in the network.


    References
    ----------
    .. [Ref2] https://freefem.org/

    """

    def __init__(self, equation=0):
        """Initialize FreeFEM.

        Parameters
        ----------
        equation : int, default 0

        Returns
        -------
        None.

        """
        self.equation = equation

        self.a1a2a3_coefficients = []

    def __prepare_script(self, network):
        """Return a FreeFEM script with `Network` geometry."""

        initialisation = textwrap.dedent(
            """
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // INITIALISATION
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            verbosity = 0;
            
            real time0=clock();
            
            // Defining the base vectors of the analytical field around a tip
            func real BaseVector(int nf, complex zf)
            {
              real result=0;
              
              if (nf%2==0) 
                result = -imag(zf^(nf/2.));
              else 
                result = real(zf^(nf/2.));
              return result;
            }
            
            // Adaptation around the tip
            func real tipfield( real[int] X, real[int] Y,real sigma,int nTips)
            {
            real err=0;
            for(int i=0;i<nTips;i++)
            	{
            		real rsq=(x-X(i))^2 + (y-Y(i))^2;
            		
            		if (rsq==0)
            			err+=1-erf(1);
            		else //if (rsq<2.*square(sigma))
            			err+=1 - 0.3*erf(1) + 0.3*erf(sqrt(rsq/(2*sigma^2)));
            		// else
            		//	err+=1;
            	}
            return err;
            }
            
            // Projection of a mesh around the tip
            func int inCircle (real x, real y, real R)
            {
                if (x^2+y^2<R^2) return 1;
                else return 0;
            }
            
            // Counting vertices in the circle around the tip
            real x0=0., y0=0.;
            func int[int] countNvAroundTips (real R, mesh Th, int nbVertices, int nbTips, real[int] X, real[int] Y)
            {
            	int[int] nvAroundTips(nbTips);
            	for(int i=0;i<nbTips;++i)
            	{
            		x0=X(i);
            		y0=Y(i);
            		int nvAroundTip = 0;
            		for (int i = 0; i < nbVertices; i++)
            			if ((x0-Th(i).x)^2 + (y0-Th(i).y)^2 < R^2) 
            				nvAroundTip += 1;		
            		nvAroundTips(i) = nvAroundTip;
            	};
            	
            	return nvAroundTips;
            }
            """
        )

        border_box = "\nreal buildTime0=clock();\n\n"
        inside_buildmesh = ""
        for i, triple in enumerate(network.box.connections_bc()):
            x0 = network.box.points[triple[0], 0]
            y0 = network.box.points[triple[0], 1]
            x1 = network.box.points[triple[1], 0]
            y1 = network.box.points[triple[1], 1]
            boundary_condition = triple[2]
            n_points = np.max((1, int(np.sqrt((x0 - x1) ** 2 + (y0 - y1) ** 2) / 2)))

            border_box = (
                border_box
                + "border box{i}(t=0, 1){{x={x0:.12g}+t*({ax:.12g});y={y0:.12g}+t*({ay:.12g}); label={bc};}}\n".format(
                    i=i, x0=x0, ax=x1 - x0, y0=y0, ay=y1 - y0, bc=boundary_condition
                )
            )

            inside_buildmesh = inside_buildmesh + " box{i}({n}) +".format(
                i=i, n=n_points
            )

        border_network = ""
        for i, branch in enumerate(network.branches):
            for j, pair in enumerate(zip(branch.points, branch.points[1:])):
                x0 = pair[0][0]
                y0 = pair[0][1]
                x1 = pair[1][0]
                y1 = pair[1][1]

                border_network = (
                    border_network
                    + "border branch{i}connection{j}(t=0, 1){{x={x0:.12g}+t*({ax:.12g});y={y0:.12g}+t*({ay:.12g}); label=1;}}\n".format(
                        i=i, j=j, x0=x0, ax=x1 - x0, y0=y0, ay=y1 - y0
                    )
                )

                inside_buildmesh = (
                    inside_buildmesh + " branch{i}connection{j}(1) +".format(i=i, j=j)
                )
        inside_buildmesh = inside_buildmesh[:-1]

        buildmesh = (
            textwrap.dedent(
                """
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // BUILDING MESH
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            """
            )
            + border_box
            + border_network
            + "\nmesh Th = buildmesh({inside_buildmesh});\n".format(
                inside_buildmesh=inside_buildmesh
            )
            + "\nreal buildTime=clock() - buildTime0;\n// plot(Th, wait=true);\n"
        )

        tip_information = textwrap.dedent(
            """
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // TIP INFORMATION
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            int nbTips={n_tips}; 
            real[int] angle(nbTips);
            real[int] X(nbTips); 
            real[int] Y(nbTips);
            """.format(
                n_tips=len(network.active_branches)
            )
        )
        for i, branch in enumerate(network.active_branches):
            tip_information = (
                tip_information
                + "\nX({j})={x:.12g};".format(j=i, x=branch.points[-1, 0])
                + "\nY({j})={y:.12g};".format(j=i, y=branch.points[-1, 1])
                + "\nangle({j})={angle:.12g};".format(
                    j=i, angle=branch.tip_angle()
                )
            )  # angle with X axis
        tip_information = tip_information + "\n"

        problem_Laplace = textwrap.dedent(
            """
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // DEFINING PROBLEM AND equation TO SOLVE
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            fespace Vh(Th,P2);
            Vh u,v;
            
            problem potential(u,v,solver=sparsesolver)=
                     int2d(Th)(dx(u)*dx(v) + dy(u)*dy(v))
                         -int1d(Th,3)(v)  // constant flux
            			 // +on(3,u=50) // constant field
            			 // -int2d(Th)(v) // rain in domain
                         +on(1,u=0);
            """
        )

        problem_Poisson = textwrap.dedent(
            """ 
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // DEFINING PROBLEM AND equation TO SOLVE
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            fespace Vh(Th,P2);
            Vh u,v;
            
            problem potential(u,v,solver=sparsesolver)=
                     int2d(Th)(dx(u)*dx(v) + dy(u)*dy(v))
                         -int2d(Th)(v) 
            			 -int1d(Th,3)(v)  // constant flux
                         +on(1,u=0);
            """
        )

        adaptmesh = textwrap.dedent(
            """
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // ADAPTING THE MESH AND SOLVING FOR THE FIELD
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////        

            // counting cells around the tips
            real R=0.01; // circle around the tip over which field is integrated
            int[int] nvAroundTips = countNvAroundTips (3.*R, Th, Th.nv, nbTips, X, Y);
            
            // First adaptation
            real firstAdaptTime0=clock();
            // Th = adaptmesh(Th,5.*tipfield(X,Y,3.*R,nbTips),nbvx=500000,nbsmooth=100,iso=true);
            Th = adaptmesh(Th,1,nbvx=500000,hmax=0.1,nbsmooth=100,iso=true,ratio=1.8);
            real firstAdaptTime=clock() - firstAdaptTime0;
            // plot(Th, wait=true);
            
            // Solving the problem for the first time
            real firstRunTime0=clock();
            potential;
            // cout<<"First solve completed."<<endl;
            real firstRunTime=clock() - firstRunTime0;
            
            // Adaptation loop
            real adaptTime0=clock();
            // cout << endl << endl << "Adaptation..." << endl;
            fespace Vh0(Th,P0);
            Vh0 h=1;
            real error=0.02;
            int adaptCounter=1;
            while(nvAroundTips.min < 250 || adaptCounter<=3)
            {
            	// cout << "Adaptation step: " << adaptCounter << ", h[].min = " << h[].min;
            	// cout << ", nvAroundTip.min = " << nvAroundTips.min << endl;
            	potential;
            	Th=adaptmesh(Th,[u, 20.*tipfield(X,Y,3.*R,nbTips)],err=error,nbvx=1000000,iso=true,ratio=2,hmin=1e-5);
            	error = 0.5*error;
            	u=u;
            	h=hTriangle; // the triangle size
            	nvAroundTips = countNvAroundTips (3.*R, Th, Th.nv, nbTips, X, Y);
            	adaptCounter++;
            }
            
            // cout << endl << "Adaptation finished." << " h[].min = " << h[].min;
            // cout << ", nvAroundTip.min = " << nvAroundTips.min << endl;
            
            // solving with adapted mesh
            potential;
            // cout << "Problem solved." << endl;
            // plot(u, wait=true, fill=true);
            
            real adaptTime=clock() - adaptTime0;
            """
        )

        integrate_a1a2a3 = textwrap.dedent(
            """
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // INTEGRATING THE FIELD TO GET a_i COEFFICIENTS
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////        
            real coeffTime0=clock();
            // cout << endl << endl << "Finding the Tip coefficients..." << endl;
            
            mesh Ph;
            real[int] a(3); // list of coefficients of the expansion
            int exponant=2; // precision of the exponential
            // ofstream freefemOutput("{file_name}");
            
            cout.precision(12);
            cout << "kopytko ";
            for(int i=0;i<nbTips;++i)
            {
                // cout << "Processing Tip " << i << " ";   
                x0=X(i);y0=Y(i);
                // cout << "(x0, y0) = (" << x0 << ", " <<y0<< "), angle = " << angle(i) << endl;
                
            	// cout << "Projecting... Th.nv = " << Th.nv;
                Ph=trunc(Th,(sqrt((x-x0)^2+(y-y0)^2) < 3*R)); 
            	// cout << ", Ph.nv = " << Ph.nv << endl;
            	
                for(int order=1; order<=a.n; ++order){ 
                    a[order-1]=
                    int2d(Ph)( u*exp(-(sqrt((x-x0)^2 + (y-y0)^2)/R)^exponant)
            		*BaseVector(order,exp(-angle(i)*1i)*( (x-x0) + (y-y0)*1i) ) ) /
                    (int2d(Ph)(exp(-(sqrt((x-x0)^2 + (y-y0)^2)/R)^exponant)
            		*square(BaseVector(order,exp(-angle(i)*1i)*( (x-x0) + (y-y0)*1i) ) )));
            		
            		cout << a[order-1] << ",";
                    // cout << "a(" << order << ") = " << a[order-1] << endl;
                }
            	// freefemOutput << Th.nv << " ";
            	// freefemOutput << Ph.nv << " ";
            	// freefemOutput << adaptCounter << " ";
            	
            	// cout << endl;
            };
            
            // cout << endl << endl << "Building mesh took: " << buildTime; 
            // cout << endl << "First adapt took: " << firstAdaptTime; 
            // cout << endl << "First run took: " << firstRunTime; 
            // cout << endl << "Adaptation took: " << adaptTime; 
            // cout << endl << "Calculating coefficients took: " << clock()- coeffTime0;
            // cout << endl << "Total time: " << clock()-time0 << endl << endl;
            """
        )

        script = initialisation + buildmesh + tip_information
        if self.equation:
            script = script + problem_Poisson
        else:
            script = script + problem_Laplace
        script = script + adaptmesh + integrate_a1a2a3

        return script

    def __run_freefem_temp(self, script):
        """Run FreeFEM from temporary file and import the a1a2a3 coefficients."""
        temporary_files = []  # to close at the end
        with NamedTemporaryFile(suffix=".edp", mode="w", delete=False) as edp_temp_file:
            edp_temp_file.write(script)
            temporary_files.append(edp_temp_file)

        cmd = [
            "FreeFem++",
            "-nw",
            "-nc",
            "-v",
            "0",
            "-f",
            "{file_name}".format(file_name=edp_temp_file.name),
        ]
        result = subprocess.run(
            args=cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if result.returncode:
            print("\nFreeFem++ failed.\n")
            print("stdout:", result.stdout.decode())
            print("stderr:", result.stderr.decode())
            
        ai_coeffs_flat = np.fromstring(result.stdout[result.stdout.find(b'kopytko')+7:], sep=",")
        self.a1a2a3_coefficients = ai_coeffs_flat.reshape(len(ai_coeffs_flat) // 3, 3)

        # close temporary files
        for tmp_file in temporary_files:
            tmp_file.close()
            os.unlink(tmp_file.name)

    def __run_freefem(self, script):
        """Run FreeFEM and import the a1a2a3 coefficients. Useful for debugging."""

        with open("script.edp", "w") as edp_temp_file:
            edp_temp_file.write(script)

        cmd = [
            "FreeFem++",
            "-nw",
            "-nc",
            "-v",
            "0",
            "-f",
            "{file_name}".format(file_name=edp_temp_file.name),
        ]
        result = subprocess.run(
            args=cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        # print("stdout:", result.stdout.decode())
        # print("stderr:", result.stderr.decode())
        if result.returncode:
            print("\nFreeFem++ failed.\n")

        ai_coeffs_flat = np.fromstring(result.stdout[result.stdout.find(b'kopytko')+7:], sep=",")
        self.a1a2a3_coefficients = ai_coeffs_flat.reshape(len(ai_coeffs_flat) // 3, 3)

    def solve_PDE(self, network):
        """Solve the PDE for the field around the network.

        Prepare a FreeFEM script, export it to a temporary file and run.
        Then, import the a1a2a3 coefficients to ``self.a1a2a3_coefficients``.

        Parameters
        ----------
        network : object of class Network
            Network around which the field will be calculated.

        Returns
        -------
        None.

        """
        script = self.__prepare_script(network)
        if system() == 'Windows':
            self.__run_freefem(script) # useful for debugging
        else:
            self.__run_freefem_temp(script)
        # with np.printoptions(formatter={'float': '{:.6e}'.format}):
            # print('a1a2a3: \n', self.a1a2a3_coefficients)


"""Handle network simulations.

Classes:
    Box
    Branch
    Network
    System
    
"""

import numpy as np
import datetime
import time
import copy
import json
import importlib.metadata

from .extending_kernels import extenders, pde_solvers, trajectory_integrators


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types.

    References
    ----------
    .. [1] https://stackoverflow.com/questions/26646362/numpy-array-is-not-json-serializable
    """

    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.int32) or isinstance(obj, np.int64): 
            return int(obj)
        return json.JSONEncoder.default(self, obj)


class Branch:
    """A class of a single branch in a network.

    Attributes
    ----------
    ID : int
        Branch ID.
    points : array
        A 2-n array with xy coordinates of the points composing the branch.
        Chronological order (tip is the last point).
    steps : array
        A 1-n array with evolution steps at which corresponding points were added.
    dR : array or 0, default 0
        A 1-2 array of tip progression ([dx, dy]).
    is_bifurcating : bool, default False
        A boolean condition if branch is bifurcating or not.
        (Based on the bifurcation_type and bifurcation_thresh from extender.)

    """

    def __init__(self, ID, points, steps):
        """Initialize Branch.

        Parameters
        ----------
        ID : int
        points : array
        steps : array

        Returns
        -------
        None.

        """
        self.ID = ID

        self.points = points  # in order of creation
        self.steps = steps  # at which step of the evolution the point was added

        self.dR = 0
        self.is_bifurcating = False

    def extend(self):
        """Add a new point to ``self.points`` (progressed tip)."""
        self.points = np.vstack((self.points, self.points[-1] + self.dR))

    def length(self):
        """Return length of the Branch."""
        return np.sum(np.linalg.norm(self.points[1:]-self.points[:-1], axis=1))

    def tip_angle(self):
        """Return the angle between the tip segment (last and penultimate point) and X axis."""
        point_penult = self.points[-2]
        point_last = self.points[-1]
        dx = point_last[0] - point_penult[0]
        dy = point_last[1] - point_penult[1]
        return np.arctan2(dy, dx)

    def points_steps(self):
        """Return a 3-n array of points and evolution steps when they were added."""
        return np.column_stack((self.points, self.steps))


class Box:
    """A class containing borders of the simulation domain.

    Attributes
    ----------
    points : array, default []
        A 2-n array with xy coordinates of the points composing the Box.
    connections : array, default []
        A 2-n array with connections between the ``points``.
    boundary_conditions : array, default []
        A 1-n array of boundary conditions \
        corresponding to links in ``connections`` list.
            - 1: absorbing BC (vanishing field)
            - 2: reflective BC (vanishing normal derivative)
            - 3: constant flux
    seeds_connectivity : array, default []
        A 2-n array of seeds connectivity.
            - 1st column: index in ``points``
            - 2nd column: outgoing branch ``ID`` 

    """

    def __init__(
        self, points=None, connections=None, boundary_conditions=None, seeds_connectivity=None
    ):
        """Initialize Box.

        Parameters
        ----------
        points : array, default []
        connections : array, default []
        boundary_conditions : array, default []
        seeds_connectivity : array, default []

        Returns
        -------
        None.

        """
        self.points = np.empty((0,2),dtype=float) if points is None else points
        self.connections = np.empty((0,2),dtype=int) if connections is None else connections
        self.boundary_conditions = np.empty((0,1),dtype=int) if boundary_conditions is None else boundary_conditions

        # 1st column: index on border
        # 2nd column: branch_id
        self.seeds_connectivity = [] if seeds_connectivity is None else seeds_connectivity
        
    def __add_points(self, points):
        self.points = np.vstack((self.points, points))

    def __add_connection(self, connections, boundary_conditions):
        self.connections = np.vstack((self.connections, connections))
        self.boundary_conditions = np.append(self.boundary_conditions, boundary_conditions)

    def connections_bc(self):
        """Return a 3-n array of connections and boundary conditions corresponding to them.
        (1st/2nd column - point indices, 3rd column - BC)
        """
        return np.column_stack((self.connections, self.boundary_conditions))

    @classmethod
    def construct(cls, initial_condition=0, **kwargs_construct):
        """Construct a Box with given initial condition.

        Parameters
        ----------
        initial_condition : int, default 0
            IC = 0 or 1. Rectangular box of dimensions ``width`` x ``height``,
            absorbing bottom wall, reflecting left and right, and:
                - IC = 0: constant flux on top (Laplacian case)
                - IC = 1: reflective top (Poissonian case)
        kwargs_construct:
            IC = 0 or 1
                seeds_x : array, default [0.1]
                    A 1-n array of x positions at the bottom boundary (y=0).
                initial_lengths : array, default [0.01]
                    A 1-n array of seeds initial lengths.
                    Length must match seeds_x or be equal to 1 
                    (then the same initial length will be set for all seeds).
                height : float, default 50.0
                    Height of the rectangular system.
                width : float, default 2.0
                    Width of the rectangular system.

        Returns
        -------
        box : Box
            An object of class Box.
        branches : list
            A list of objects of class Branch.


        """
        # Labels for boundary conditions
        DIRICHLET = 1
        NEUMANN = 2
        CONSTANT_FLUX = 3

        # Build a box
        box = cls()

        # Rectangular box of specified width and height
        # IC==0: Constant flux on top (Laplace)
        # IC==1: Neumann on top (Poisson)
        if initial_condition == 0 or initial_condition == 1:
            options_construct = {
                "seeds_x": [0.1],
                "initial_lengths": [0.01],
                "height": 50.0,
                "width": 2.0,
            }
            options_construct.update(kwargs_construct)
            box.__add_points(
                [
                    [options_construct["width"], 0],
                    [options_construct["width"], options_construct["height"]],
                    [0, options_construct["height"]],
                    [0, 0],
                ]
            )
            box.seeds_connectivity = np.column_stack(
                (
                    len(box.points) +
                    np.arange(len(options_construct["seeds_x"])),
                    np.arange(len(options_construct["seeds_x"])),
                )
            )
            box.__add_points(
                np.vstack(
                    [
                        options_construct["seeds_x"],
                        np.zeros(len(options_construct["seeds_x"])),
                    ]
                ).T
            )

            connections_to_add = np.vstack(
                [np.arange(len(box.points)), np.roll(
                    np.arange(len(box.points)), -1)]
            ).T
            box.__add_connection(
                connections_to_add,
                boundary_conditions=DIRICHLET
                * np.ones(len(connections_to_add), dtype=int),
            )

            # right, left, top Neumann:
            box.boundary_conditions[0:3] = NEUMANN
            # or top constant flux:
            if initial_condition == 0:
                box.boundary_conditions[1] = CONSTANT_FLUX

            # Creating initial branches
            branches = []
            if len(options_construct["initial_lengths"]) == 1:
                options_construct["initial_lengths"] = (
                    np.ones(len(options_construct["seeds_x"]))
                    * options_construct["initial_lengths"][0]
                )
            for i, x in enumerate(options_construct["seeds_x"]):
                branches.append(
                    Branch(
                        ID=i,
                        points=np.array(
                            [[x, 0], [x, options_construct["initial_lengths"][i]]]
                        ),
                        steps=np.array([0, 0]),
                    )
                )

        return box, branches


class Network:
    """A class of network with its branches and containing box.

    Attributes
    ----------
    box : Box
        An object of class Box.
    branches : list, default []
        A list of all branches (objects of class Branch) composing the network.
    active_branches : list, default []
        A list of all branches that still extend.
    sleeping_branches : list, default []
        A list of all branches without enough
        flux to move (may revive in the Poisson case - TO DO).
    branch_connectivity : array, default []
        A 2-n array with connections between the branches
        (noted with branch IDs).

    """

    def __init__(
        self,
        box,
        branches=None,
        active_branches=None,
        sleeping_branches=None,
        branch_connectivity=None,
    ):
        """Initialize Network.

        Parameters
        ----------
        box : Box
        branches : list, default []
        active_branches : list, default []
        sleeping_branches : list, default []
        branch_connectivity : array, default []

        Returns
        -------
        None.

        """

        self.box = box

        # all branches (to construct mesh): moving + sleeping + branches inside the tree
        self.branches = [] if branches is None else branches
        self.active_branches = [] if active_branches is None else active_branches  # moving branches (to extend)
        # branches without enough flux to move (may revive in the Poisson case)
        self.sleeping_branches = [] if sleeping_branches is None else sleeping_branches

        self.branch_connectivity = [] if branch_connectivity is None else branch_connectivity

    def copy(self):
        """Return a deepcopy of the Network."""
        return copy.deepcopy(self)

    def height_and_length(self):
        """Return network height (max y coordinate) and total length of the branches."""
        ruler = 0
        height = 0
        for branch in self.branches:
            ruler = ruler + branch.length()
            height = np.max((height, np.max(branch.points[:, 1])))
        return height, ruler

    def __add_connection(self, connection):
        """Add connection to self.branch_connectivity."""
        if not len(self.branch_connectivity):
            self.branch_connectivity = connection
        else:
            self.branch_connectivity = np.vstack(
                (self.branch_connectivity, connection))

    def move_tips(self, step):
        """Move tips (with bifurcations and killing).

        Parameters
        ----------
        step : int, default 0
            The current evolution step.

        Returns
        -------
        None.

        """

        # shallow copy of active_branches (creates new list instance, but the elements are still the same)
        branches_to_iterate = self.active_branches.copy()
        for branch in branches_to_iterate:
            if branch.is_bifurcating:
                print("! Branch {ID} bifurcated !".format(ID=branch.ID))
                max_branch_id = len(self.branches) - 1

                for i, dR in enumerate(branch.dR):
                    points = np.array(
                        [branch.points[-1], branch.points[-1] + dR])
                    branch_new = Branch(
                        ID=max_branch_id + i + 1,
                        points=points,
                        steps=np.array([step - 1]),
                    )
                    self.branches.append(branch_new)
                    self.active_branches.append(branch_new)
                    self.__add_connection(
                        np.array([branch.ID, branch_new.ID]))
                self.active_branches.remove(branch)
            else:
                branch.extend()


class System:
    """A class containing all the elements to run a network simulation.

    Attributes
    ----------
    network : Network
        An object of class Network.
    extender : Extender
        An object of one of the classes from reticuler.extending_kernels.extenders.
    trajectory_integrator : TrajectoryIntegrator
        One of the classes from reticuler.extending_kernels.trajectory_integrators.
    growth_thresh_type : int, default 0
        Type of growth threshold.
            - 0: max step
            - 1: height
            - 2: network length
    growth_thresh : float, default 5
        A value of growth threshold. The simulation is stopped, when it's reached.
    growth_gauges : array, default array([0.,0.,0.])
        A 1-3 array with growth gauges (max step, height, network length).
    dump_every : int, default 1
        Dumps the results every ``dump_every`` steps.
    exp_name: str, default ''
        Path to a file, where the results will be stored.

    """

    def __init__(
        self,
        network,
        extender,
        trajectory_integrator,
        growth_thresh_type=0,
        growth_thresh=5,
        growth_gauges=None,
        dump_every=1,
        exp_name="",
    ):
        """Initialize System.

        Parameters
        ----------
        network : Network
        extender : Extender
        trajectory_integrator : function
        growth_gauges : array, default array([0.,0.,0.,0.])
        growth_thresh_type : int, default 0
        growth_thresh : float, default 5
        dump_every : int, default 1
        exp_name: str, default ''

        Returns
        -------
        None.

        """
        self.network = network
        self.extender = extender
        self.trajectory_integrator = trajectory_integrator

        # Growth limits:
        # 0: max step, 1: max height
        # 2: max length 3: max time
        self.growth_gauges = np.zeros(4) if growth_gauges is None else growth_gauges
        self.growth_thresh_type = growth_thresh_type
        self.growth_thresh = growth_thresh

        self.dump_every = dump_every
        self.exp_name = exp_name
    
    def copy(self):
        """Return a deepcopy of the Network."""
        return copy.deepcopy(self)

    def export_json(self):
        """Export all the information to 'self.exp_name'+'.json'."""
        growth_type_legend = ["max step",
                              "max height", "max length", "max time"]
        export_general = {
            "reticuler_version": importlib.metadata.version("reticuler"),
            "exp_name": self.exp_name,
            "growth": {
                "threshold_type": growth_type_legend[self.growth_thresh_type],
                "threshold": self.growth_thresh,
                "growth_gauges": {
                    "number_of_steps": self.growth_gauges[0],
                    "height": self.growth_gauges[1],
                    "network_length": self.growth_gauges[2],
                    "time": self.growth_gauges[3],
                },
                "dump_every": self.dump_every,
            },
        }

        if type(self.extender).__name__ == "Streamline":
            if type(self.trajectory_integrator).__name__ == "ModifiedEulerMethod":
                export_trajectory_integrator = {
                    "type": type(self.trajectory_integrator).__name__,
                    "max_approximation_step": self.trajectory_integrator.max_approximation_step,
                }
            if type(self.extender.pde_solver).__name__ == "FreeFEM":
                equation_legend = ["Laplace", "Poisson"]
                export_solver = {
                    "type": type(self.extender.pde_solver).__name__,
                    "equation": equation_legend[self.extender.pde_solver.equation],
                }

            bifurcation_type_legend = [
                "no bifurcations", "a1", "a3/a1", "random"]
            export_extender = {
                "type": type(self.extender).__name__,
                "eta": self.extender.eta,
                "ds": self.extender.ds,
                "bifurcations": {
                    "type": bifurcation_type_legend[
                        self.extender.bifurcation_type
                    ],
                    "threshold": self.extender.bifurcation_thresh,
                    "angle": self.extender.bifurcation_angle,
                },
                "inflow_thresh": self.extender.inflow_thresh,
                "distance_from_bif_thresh": self.extender.distance_from_bif_thresh,
                "pde_solver": {**export_solver},
            }

            export_extending_kernel = {
                "extending_kernel": {
                    "trajectory_integrator": {**export_trajectory_integrator},
                    "extender": {**export_extender},
                }
            }

        export_branches = {}
        for branch in self.network.branches[::-1]:
            if branch in self.network.active_branches:
                state = "active"
            elif branch in self.network.sleeping_branches:
                state = "sleeping"
            else:
                state = "dead"
            branch_dict = {
                branch.ID: {
                    "ID": branch.ID,
                    "state": state,
                    "points_and_steps": branch.points_steps(),
                }
            }
            export_branches = export_branches | branch_dict
        export_network = {
            "network": {
                "description": "Geometry of the system: box and branches.",
                "box": {
                    "description": "Border geometry. Points should be in a counterclokwise order. Connections and boundary conditions (BC) -> 1st/2nd columns: point IDs, 3rd column: BC. Seeds connectivity -> 1st column: index on border, 2nd column: branch ID.",
                    "points": self.network.box.points,
                    "connections_and_bc": self.network.box.connections_bc(),
                    "seeds_connectivity": self.network.box.seeds_connectivity,
                },
                "branch_connectivity": self.network.branch_connectivity,
                "branches": {**export_branches},
            }
        }

        to_export = export_general | export_extending_kernel | export_network
        with open(self.exp_name + ".json", "w", encoding="utf-8") as f:
            json.dump(to_export, f, ensure_ascii=False,
                      indent=4, cls=NumpyEncoder)

    @classmethod
    def import_json(cls, input_file):
        """Construct an instance of class System based on the imported .json file.

        Parameters
        ----------
        input_file : path
            Name of the experiment location. Extension '.json' will be added.

        Returns
        -------
        system : object of class System

        """
        with open(input_file + ".json", "r") as f:
            json_load = json.load(f)

        # Branches
        branches = []
        active_branches = []
        sleeping_branches = []
        for i in reversed(list(json_load["network"]["branches"].keys())):
            json_branch = json_load["network"]["branches"][i]
            points_steps = np.asarray(json_branch["points_and_steps"])
            branch = Branch(
                ID=json_branch["ID"],
                points=points_steps[:, :2],
                steps=np.array(points_steps[:, 2], dtype=int),
            )

            branches.append(branch)
            if json_branch["state"] == "active" or json_branch["state"] == "sleeping":
                active_branches.append(branch)
            # elif json_branch["state"] == "sleeping":
            #     sleeping_branches.append(branch)

        # Box
        json_box = json_load["network"]["box"]
        connections_bc = np.asarray(json_box["connections_and_bc"])
        box = Box(
            points=np.asarray(json_box["points"]),
            connections=connections_bc[:, :2],
            boundary_conditions=connections_bc[:, 2],
            seeds_connectivity=np.asarray(json_box["seeds_connectivity"]),
        )
        # Network
        branch_connectivity = np.asarray(
            json_load["network"]["branch_connectivity"])
        network = Network(
            box=box,
            branches=branches,
            active_branches=active_branches,
            sleeping_branches=sleeping_branches,
            branch_connectivity=branch_connectivity,
        )

        # Trajectory integrator
        json_trajectory_integrator = json_load["extending_kernel"][
            "trajectory_integrator"
        ]
        if json_trajectory_integrator["type"] == "ModifiedEulerMethod":
            max_approximation_step = json_trajectory_integrator["max_approximation_step"]
            trajectory_integrator = trajectory_integrators.ModifiedEulerMethod(
                max_approximation_step=max_approximation_step)

        # Solver
        json_solver = json_load["extending_kernel"]["extender"]["pde_solver"]
        if json_solver["type"] == "FreeFEM":
            equation_legend = ["Laplace", "Poisson"]
            equation = equation_legend.index(json_solver["equation"])
            pde_solver = pde_solvers.FreeFEM(equation=equation)

        # Extender
        json_extender = json_load["extending_kernel"]["extender"]
        if json_extender["type"] == "Streamline":
            json_bifurcation = json_extender["bifurcations"]
            bifurcation_type_legend = [
                "no bifurcations", "a1", "a3/a1", "random"]
            bifurcation_type = bifurcation_type_legend.index(
                json_bifurcation["type"])
            extender = extenders.Streamline(
                pde_solver=pde_solver,
                eta=json_extender["eta"],
                ds=json_extender["ds"],
                bifurcation_type=bifurcation_type,
                bifurcation_thresh=json_bifurcation["threshold"],
                bifurcation_angle=json_bifurcation["angle"],
                inflow_thresh=json_extender["inflow_thresh"],
            )
            extender.distance_from_bif_thresh = json_extender[
                "distance_from_bif_thresh"
            ]

        # General
        json_growth = json_load["growth"]
        growth_type_legend = ["max step",
                              "max height", "max length", "max time"]
        growth_thresh_type = growth_type_legend.index(
            json_growth["threshold_type"])
        growth_thresh = json_growth["threshold"]
        dump_every = json_growth["dump_every"]

        json_growth_gauges = json_growth["growth_gauges"]
        growth_gauges = np.array(
            [
                json_growth_gauges["number_of_steps"],
                json_growth_gauges["height"],
                json_growth_gauges["network_length"],
                json_growth_gauges["time"],
            ]
        )

        system = cls(
            network=network,
            extender=extender,
            trajectory_integrator=trajectory_integrator,
            growth_gauges=growth_gauges,
            growth_thresh_type=growth_thresh_type,
            growth_thresh=growth_thresh,
            dump_every=dump_every,
            exp_name=input_file,
        )

        return system

    def __update_growth_gauges(self, dt):
        """Update growth gauges."""
        self.growth_gauges[1], self.growth_gauges[2] = self.network.height_and_length(
        )
        self.growth_gauges[3] = self.growth_gauges[3] + dt

        print("Active branches: {n:d}".format(
            n=len(self.network.active_branches)))
        print("Network height: {h:.3f}".format(h=self.growth_gauges[1]))
        print("Network length: {l:.3f}".format(l=self.growth_gauges[2]))
        print("Evolution time: {t:.3f}".format(t=self.growth_gauges[3]))

        for branch in self.network.active_branches:
            branch.steps = np.append(branch.steps, self.growth_gauges[0])
            # branch.timestamps = np.append( branch.timestamps, self.growth_gauges[3] )

    def evolve(self):
        """Run the simulation.

        Run the simulation in a while loop until ``self.growth_thresh`` is not reached.
        Print real time that the simulation took.

        Returns
        -------
        None.

        """
        start_clock = time.time()
        while self.growth_gauges[self.growth_thresh_type] < self.growth_thresh:
            self.growth_gauges[0] = self.growth_gauges[0] + 1
            print(
                "\n-------------------   Growth step: {step:.0f}   -------------------\n".format(
                    step=self.growth_gauges[0]
                )
            )
            print("Date and time: ", datetime.datetime.now())

            dt, _ = self.trajectory_integrator.integrate(
                extender=self.extender, network=self.network)
            self.network.move_tips(step=self.growth_gauges[0])
            self.__update_growth_gauges(dt)

            if not self.growth_gauges[0] % self.dump_every:
                self.export_json()

        self.export_json()

        print(
            "\n End of the simulation. Time: {clock:.2f}s".format(
                clock=time.time() - start_clock
            )
        )



"""Trimmers to execute backward step.

Classes:
    BackwardModifiedEulerMethod

"""

import numpy as np
import copy


class BackwardModifiedEulerMethod:
    """A class to integrate tip trajectories with reversed modified Euler's method and streamline algorithm.

    Reversed modified Euler's method:

    .. math:: x(n-1) = x(n) - dt \cdot 0.5 \cdot (v[x(n)] + v[x(n-1)]).

    Attributes
    ----------
    pde_solver : PDESolver
    eta : float, default 1.0
        The growth exponent (v=a1**eta).
    ds : float, default 0.01
        A distance over which the fastest branch in the network
        will move in each timestep.    
    max_approximation_step : int, default 3
        Number of approximation steps:
            - 0:  explicit Euler's method
            - 1:  Heuns's method
            - >1: modified Euler's method    
    inflow_thresh : float, default 0.05
        Threshold to put asleep the tips with less than ``inflow_thresh``
        of max flux/velocity.

    References
    ----------
    .. [Ref1] "Through history to growth dynamics: backward evolution of spatial networks",
            S. Żukowski, P. Morawiecki, H. Seybold, P. Szymczak, Sci Rep 12, 20407 (2022). 
            https://doi.org/10.1038/s41598-022-24656-x
    .. [Ref3] https://en.wikipedia.org/wiki/Trapezoidal_rule_(differential_equations)

    """
    def __init__(
        self,
        pde_solver,
        eta=0.0,
        ds=0.01,
        max_approximation_step=3,
        inflow_thresh = 0.05,
    ):
        """Initialize BackwardModifiedEulerMethod.

        Parameters
        ----------
        pde_solver : PDESolver
        eta : float, default 1.0
        ds : float, default 0.01
        max_approximation_step : int, default 3
        inflow_thresh : float, default 0.05

        Returns
        -------
        None.

        """
        self.pde_solver = pde_solver

        self.eta = eta
        self.ds = ds

        self.max_approximation_step = max_approximation_step
        
        # less than `inflow_thresh` of max flux/velocity puts branches asleep
        self.inflow_thresh = inflow_thresh
        
    def __check_moving_conditions(self, network):
        """Check moving conditions."""
        a1 = self.pde_solver.a1a2a3_coefficients[:, 0]
        max_a1 = np.max(a1)
        # (first condition for low eta, second for high)
        are_moving = np.logical_and(a1/max_a1 > self.inflow_thresh,
                                   (a1/max_a1)**self.eta > self.inflow_thresh)
        # shallow copy of active_branches (creates new list instance, but the elements are still the same)
        branches_to_iterate = network.active_branches.copy()
        for i, branch in enumerate(branches_to_iterate):
            if not are_moving[i]:
                network.sleeping_branches.append(branch)
                network.active_branches.remove(branch)
                # print("! Branch {ID} is sleeping !".format(ID=branch.ID))                
        return are_moving

    def __explicit_network_trim(self, test_network, drs_0, backward_branches, backward_bifurcations, BEA_step):
        """Trim branches with given ``drs``."""
        min_distance = 0.0001  # avoid two nodes very close to each other after trimming

        drs = drs_0.copy()
        branches_to_iterate = test_network.active_branches.copy()
        are_living_after_bif = np.ones(len(branches_to_iterate), dtype=bool)
        for i, forward_branch in enumerate(branches_to_iterate):
            
            while drs[i] > 0:
                backward_branch = backward_branches[forward_branch.ID]
                # check the distances from the tip to bifurcation point
                segment_lengths = np.linalg.norm(
                    forward_branch.points[1:]-forward_branch.points[:-1], axis=1)
                length_from_tip = np.cumsum(segment_lengths[::-1])
                       
                # bifurcation not reached
                if drs[i]+min_distance < length_from_tip[-1]:
                    are_living_after_bif[i] = True
                    
                    # how many points to remove
                    to_remove = np.sum(drs[i]+min_distance > length_from_tip)
                    
                    if to_remove>0:
                        forward_branch.points = forward_branch.points[:-to_remove]
                        forward_branch.steps = forward_branch.steps[:-to_remove]
                        drs[i] = drs[i] - length_from_tip[to_remove-1]
                    
                    # shifting the last point
                    if drs[i]!=0:
                        tip_versor = forward_branch.points[-1] - forward_branch.points[-2]
                        tip_versor = tip_versor / np.linalg.norm(tip_versor)
                        forward_branch.points[-1] = forward_branch.points[-1] - tip_versor * drs[i]
                        backward_branch.add_point(forward_branch.points[-1], len(forward_branch.points)-1, BEA_step) 
                        drs[i] = 0
                # bifurcation reached
                else:
                    drs[i] = drs[i] - length_from_tip[-1]
                    backward_branch.add_point(forward_branch.points[0], 0, BEA_step) 
                                    
                    mask = test_network.branch_connectivity[:,1]!=forward_branch.ID
                    test_network.branch_connectivity = test_network.branch_connectivity[mask,...]
                    test_network.branches.remove(forward_branch)
                    ind_branch = test_network.active_branches.index(forward_branch)
                    
                    # if we reach the border
                    if backward_branch.mother_ID==-1:
                        mask = test_network.box.seeds_connectivity[:,1]!=forward_branch.ID
                        test_network.box.seeds_connectivity = test_network.box.seeds_connectivity[mask,...]
                        test_network.active_branches.pop(ind_branch)
                    # bifurcation point
                    else:
                        ind_bifurcation = backward_bifurcations.mother_IDs==backward_branch.mother_ID
                        if backward_bifurcations.flags[ind_bifurcation] == 0:
                            print("! Branch {ID} reached bifurcation {bifID} ! (1st one)".format(ID=backward_branch.ID, bifID=backward_branch.mother_ID))
                            backward_bifurcations.flags[ind_bifurcation] = 1
                            are_living_after_bif[i] = False
                            # branched is later poped from initial_network (back in the trim method)
                            test_network.active_branches.pop(ind_branch)
                            
                            # total length of the remaining branches connected to the mother branch
                            remaining_length, remaining_IDs = self.__calculate_remaining_length(test_network, backward_branch.mother_ID)
                            inds_remaining = [i for i, b in enumerate(branches_to_iterate) if b.ID in remaining_IDs]
                            
                            # if drs[i]>0 when tip(i) reaches the bifPoint, then dt was too big;
                            # we compansate the effect by adding the 'virtual length'
                            virtual_length = drs[i] / drs_0[i] * np.sum(drs_0[inds_remaining])
                            
                            # if remaining branches weren't moved yet we have to subtract their drs
                            length_mismatch = remaining_length + virtual_length - np.sum(drs[inds_remaining])
                            
                            # if length_mismatch < 0 then we have to reverse the above line and calculate virtual_length for the remaining tip
                            if length_mismatch < 0:
                                virtual_length = (np.sum(drs[inds_remaining])-remaining_length) / np.sum(drs_0[inds_remaining]) * drs_0[i]
                                length_mismatch = virtual_length - drs[i]
                            
                            backward_bifurcations.length_mismatch[ind_bifurcation] = length_mismatch
                            drs[i] = 0 # stops while loop
                            
                        elif backward_bifurcations.flags[ind_bifurcation] == 1:
                            print("! Branch {ID} reached bifurcation {bifID} ! (2nd one)".format(ID=backward_branch.ID, bifID=backward_branch.mother_ID))
                            backward_bifurcations.flags[ind_bifurcation] = 2
                            are_living_after_bif[i] = True
                            
                            mother_branch = [b for b in test_network.branches if b.ID==backward_branch.mother_ID][0]
                            forward_branch = mother_branch
                            test_network.active_branches[ind_branch] = mother_branch
                        
        return are_living_after_bif
    
    
    def __calculate_remaining_length(self, network, mother_ID):
        to_check = network.branch_connectivity[network.branch_connectivity[:,0]==mother_ID,1]
        branch_IDs = np.array([b.ID for b in network.branches])
        remaining_length = 0
        remaining_IDs = []
        while to_check.size:
            next_ID = to_check[0]
            remaining_length = remaining_length + network.branches[np.where(branch_IDs==next_ID)[0][0]].length()
            
            to_add = network.branch_connectivity[network.branch_connectivity[:,0]==next_ID,1]
            if to_add.size:
                to_check = np.concatenate((to_check, to_add))
                remaining_IDs.append(next_ID)
            else:
                remaining_IDs.append(next_ID)
            to_check = to_check[1:]
        return remaining_length, remaining_IDs
    
    def trim(self, network, backward_branches_0, backward_bifurcations_0, BEA_step):
        """Perform backward step with the reversed modified Euler's method and the streamline algorithm.
    
        Parameters
        ----------
        network : Network
            An object of class Network.
        backward_branches_0 : list
            A list of all backward branches.
        backward_bifurcations_0 : BackwardBifurcations
            An object of class BackwardBifurcations.
        BEA_step : int
            Current step of the BEA.            
    
        Returns
        -------
        initial_network : Network
            The initial network with updated active_branches list (removed branches: (i) aren't moving, (ii) reached bifurcation)
        test_network : Network
            A network after backward evolution step.
        backward_branches : list
            An updated list of all backward branches.
        backward_bifurcations : BackwardBifurcations
            An updated object of class BackwardBifurcations.
        dt : float
            Time of (backward) growth at the current BEA step.
        
        Returns
        -------
        None.
                
        """
        # activate sleeping branches        
        network.active_branches = network.sleeping_branches + network.active_branches
        network.sleeping_branches = []
        
        # v[x(n)]: finding velocity at the starting point
        self.pde_solver.solve_PDE(network)
        velocity_0 = self.pde_solver.a1a2a3_coefficients[:, 0]**self.eta
        dt_0 = self.ds / np.max(velocity_0)
        drs_0 = dt_0 * velocity_0

        # are moving?
        are_moving = self.__check_moving_conditions(network) # here we put branches to sleep
        drs_0 = drs_0[are_moving]
        initial_network = network.copy()
        
        # x(n-1): trimming drs_0 from test_network
        backward_branches = copy.deepcopy(backward_branches_0)
        backward_bifurcations = copy.deepcopy(backward_bifurcations_0)
        test_network = network.copy()
        are_living_after_bif = self.__explicit_network_trim(test_network, drs_0, backward_branches, backward_bifurcations, BEA_step)
        if test_network.active_branches:
            drs = np.zeros_like(drs_0)   
            approximation_step = 0
            # APPROXIMATION LOOP
            while approximation_step < self.max_approximation_step and not (~are_living_after_bif).all():
                approximation_step = approximation_step + 1

                # v[x(n-1)]: finding velocity at the x(n-1) point
                self.pde_solver.solve_PDE(test_network)
                velocity_1 = self.pde_solver.a1a2a3_coefficients[:, 0]**self.eta
                drs_1 = dt_0 * velocity_1

                # 0.5 * dt * (v[x(n)] + v[x(n-1)]): average drs
                # drs[np.logical_not(are_living_after_bif)] = drs_0[np.logical_not(are_living_after_bif)] / 2
                # !!! dead tips after bif don't move, so dt*(v[x(n)] + v[x(n-1)])/2 = dt*v[x(n-1)]/2 - it was commented in matlab code?
                drs[are_living_after_bif] = (drs_0[are_living_after_bif] + drs_1) / 2
                velocity = drs / dt_0
                dt = self.ds / np.max(velocity)
                drs = dt * velocity

                # improved x(n-1)
                backward_branches = copy.deepcopy(backward_branches_0)
                backward_bifurcations = copy.deepcopy(backward_bifurcations_0)
                test_network = network.copy()
                are_living_after_bif = self.__explicit_network_trim(test_network, drs, backward_branches, backward_bifurcations, BEA_step)
                
            initial_network.active_branches = [b for i, b in enumerate(network.active_branches) if are_living_after_bif[i]]
        else:
            dt = 0
        
        return initial_network, test_network, backward_branches, backward_bifurcations, dt        
        



"""Handle the Backward Evolution Algorithm.

Classes:
    Metrics
    Network
    System
    
"""

import time
import datetime
import numpy as np
import importlib.metadata
import json

from reticuler.system import NumpyEncoder
from reticuler.backward_evolution import trimmers


class BackwardBranch:
    """A class of a single branch in a network used to collect BEA metrics.

    Attributes
    ----------
    ID : int
        Branch ID.
    mother_ID : int
        ID of the mother branch (branch from which `self` bifrucated). 
        If `self` is connected to the border mother_ID=-1.
    points : array
        A 2-n array with xy coordinates of the points composing the branch.
        Chronological order of growth (tip is the last point).
    nums_left : array
        A 1-n array with the number of points left in the forward branch at the current BEA step.
        Used to clip reconstruct the state of the network during BEA.
    steps : array
        A 1-n array with BEA steps at which corresponding points were added.
    a1a2a3_coefficients : array
        A 3-n array with a1,a2,a3 coefficients in a point after backward step.
    overshoot : array
        A 1-n array with overshoot after virtual forward step.
    angular_deflection : array
        A 1-n array with angular deviation after virtual forward step.

    """

    def __init__(self, ID, mother_ID):
        """Initialize BackwardBranch.

        Parameters
        ----------
        ID : int
        points : array
        steps : array

        Returns
        -------
        None.

        """
        self.ID = ID
        self.mother_ID = mother_ID

        self.points = np.empty((0,2), dtype=float)  # in order of trimming
        self.nums_left = [] # how many points are left in the forward branch
        self.steps = [] # at which step of the BEA the point was reached

        # arrays with BEA metrics are two elements shorter than points/steps
        # we don't collect metrics at the first and the last point of the initial branch
        self.a1a2a3_coefficients = np.empty((0,3), dtype=float)
        self.overshoot = []
        self.angular_deflection = []
        
    def add_point(self, point, ind, BEA_step):
        """Add a new point to ``self.points`` (position of the tip after trimming)."""
        self.points = np.vstack((self.points, point))
        self.nums_left = np.append(self.nums_left, ind)
        self.steps = np.append(self.steps, BEA_step)

    def points_numsleft_steps(self):
        """Return a 3-n array of points, number of points left in the forward branch and BEA steps when they were added."""
        return np.column_stack((self.points, self.nums_left, self.steps))


class BackwardBifurcations:
    """A class containing BEA metrics at all bifurcation points.

    Attributes
    ----------
    mother_IDs : array
        A 1-n array with all IDs of the mother branchs of the bifurcations.
    a1a2a3_coefficients : array
        A 3-n array with a1,a2,a3 coefficients in each bifurcation point 
        after two branches reached it.
    length_mismatch : array
        A 1-n array with length mismatch in each bifurcation point
        after one of the branches reached it.
    flags : array
        A 1-n array of flags:
            - 0: bifurcation not reached
            - 1: reached once
            - 2: reached twice
            - 3: a1a2a3 coefficients stored

    """
    def __init__(self, mother_IDs):
        """Initialize BackwardBifurcations.

        Parameters
        ----------
        mother_IDs : array

        Returns
        -------
        None.

        """
        self.mother_IDs = mother_IDs
        
        self.a1a2a3_coefficients = np.zeros((len(mother_IDs), 3))
        self.length_mismatch = np.zeros((len(mother_IDs), 1))
        
        # flag: 0-not reached, 1-reached once, 2-reached twice, 3-data saved
        self.flags = np.zeros(len(mother_IDs), dtype=int)
    
    def bif_info(self):
        """Return a 6-n array with bifurcation info (mother_ID, a1, a2, a3, length mismatch, flag)."""
        return np.column_stack((self.mother_IDs, self.a1a2a3_coefficients, self.length_mismatch, self.flags))

class BackwardSystem:
    """A class containing all the elements to run the Backward Evolution Algorithm.

    Attributes
    ----------
    system : System
        An object of class System.
    trimmers : Trimmer
        An object of one of the classes from reticuler.BEA.trimmers.
    backward_branches : list
        A list of all backward branches.
    backward_bifurcations : BackwardBifurcations
        An object of class BackwardBifurcations.           
    BEA_step : int
        Current step of the BEA.
    BEA_step_thresh : int, default 1e4
        A value of BEA step threshold. The simulation is stopped, when it's reached.
    back_forth_steps_thresh : int, default 1
        Number of backward (and later forward) steps in one BEA step.
    dump_every : int, default 1
        Dumps the results every ``dump_every`` steps.
    exp_name: str, default  ``system.exp_name``+'_back'
        Path to a file, where the results will be stored.

    """

    def __init__(
        self,
        system,
        trimmer,
        BEA_step_thresh=1e4,
        back_forth_steps_thresh=1,
        dump_every=1,
    ):
        """Initialize System.

        Parameters
        ----------
        system : System
        trimmers : Trimmer
        back_forth_steps_thresh : int, default 1
        BEA_step_thresh : int, default 1e4
        dump_every : int, default 1

        Returns
        -------
        None.

        """
        self.system = system
        self.trimmer = trimmer

        self.back_forth_steps_thresh = back_forth_steps_thresh
        self.BEA_step = 0
        self.BEA_step_thresh = BEA_step_thresh

        # forward/backward branch ID is its index in backward_branches
        self.backward_branches, self.backward_bifurcations = self.import_branches(self.system.network)
        self.system.extender.eta = self.trimmer.eta
        self.system.extender.bifurcation_type = 0
        self.system.extender.inflow_thresh = 0
        
        self.dump_every = dump_every
        self.exp_name = system.exp_name+'_back'
        
    @classmethod
    def import_json(cls, input_file, system):
        """Construct an instance of class BackwardSystem based on the imported .json file.

        Parameters
        ----------
        input_file : path
            Name of the experiment location. Extension '.json' will be added.

        Returns
        -------
        backward_system : object of class BackwardSystem

        """
        with open(input_file + ".json", "r") as f:
            json_load = json.load(f)

        # Backward branches
        backward_branches = []
        for i in reversed(list(json_load["backward_branches"].keys())):
            json_branch = json_load["backward_branches"][i]
            backward_branch = BackwardBranch(
                ID=json_branch["ID"],
                mother_ID=json_branch["mother_ID"],
            )
            
            points_numsleft_steps = np.asarray(json_branch["points_numsleft_steps"])
            if points_numsleft_steps.size:
                backward_branch.points = points_numsleft_steps[:, :2]
                backward_branch.nums_left = np.array(points_numsleft_steps[:, 2], dtype=int)
                backward_branch.steps = np.array(points_numsleft_steps[:, 3], dtype=int)
                backward_branch.a1a2a3_coefficients = np.asarray(json_branch["a1a2a3_coefficients"])
                backward_branch.overshoot = np.asarray(json_branch["overshoot"])
                backward_branch.angular_deflection = np.asarray(json_branch["angular_deflection"])

            backward_branches.append(backward_branch)

        # BackwardBifurcations
        bif_info = np.asarray(json_load["backward_bifurcations"]["bif_info"])
        backward_bifurcations = BackwardBifurcations(np.array(bif_info[:,0],dtype=int))
        backward_bifurcations.a1a2a3_coefficients = bif_info[:, 1:4]
        backward_bifurcations.length_mismatch = bif_info[:, 4]
        backward_bifurcations.flags = np.array(bif_info[:,5], dtype=int)

        # Trimmer
        json_trimmer = json_load["trimmer"]
        if json_trimmer["type"] == "BackwardModifiedEulerMethod":
            trimmer = trimmers.BackwardModifiedEulerMethod(
                pde_solver=system.extender.pde_solver,
                eta=json_trimmer["eta"],
                ds=json_trimmer["ds"],
                max_approximation_step=json_trimmer["max_approximation_step"],
                inflow_thresh=json_trimmer["inflow_thresh"],
            )

        # General
        json_BEA_parameters = json_load["BEA_parameters"]
        BEA_step = json_BEA_parameters["BEA_step"]
        BEA_step_thresh = json_BEA_parameters["BEA_step_thresh"]
        back_forth_steps_thresh = json_BEA_parameters["back_forth_steps_thresh"]
        dump_every = json_BEA_parameters["dump_every"]

        backward_system = cls(
            system=system,
            trimmer=trimmer,
            BEA_step_thresh=BEA_step_thresh,
            back_forth_steps_thresh=back_forth_steps_thresh,
            dump_every=dump_every,
        )
        backward_system.BEA_step = BEA_step
        backward_system.backward_branches = backward_branches
        backward_system.backward_bifurcations = backward_bifurcations

        return backward_system

    def export_json(self):
        """Export all the information to 'self.exp_name'+'.json'."""
        
        export_general = {
            "reticuler_version": importlib.metadata.version("reticuler"),
            "exp_name": self.exp_name,
            "BEA_parameters": {
                "back_forth_steps_thresh": self.back_forth_steps_thresh,
                "BEA_step": self.BEA_step,
                "BEA_step_thresh": self.BEA_step_thresh,
                "dump_every": self.dump_every,
            },
        }
        
        if type(self.trimmer.pde_solver).__name__ == "FreeFEM":
            equation_legend = ["Laplace", "Poisson"]
            export_solver = {
                "type": type(self.trimmer.pde_solver).__name__,
                "equation": equation_legend[self.trimmer.pde_solver.equation],
            }
        
        if type(self.trimmer).__name__ == "BackwardModifiedEulerMethod":
            export_trimmer = {
                    "trimmer": {
                        "type": type(self.trimmer).__name__,
                        "eta": self.trimmer.eta,
                        "ds": self.trimmer.ds,
                        "max_approximation_step": self.trimmer.max_approximation_step,
                        "inflow_thresh": self.trimmer.inflow_thresh,
                        "pde_solver": {**export_solver},
                    }
                }

        export_backward_bifurcations = {
            "backward_bifurcations": {
                "description": "Information gathered in the bifurcation points: mother_ID, a1, a2, a3, length mismatch, flag",
                "bif_info": self.backward_bifurcations.bif_info(),
            }
        }
        
        export_backward_branches = {}
        for branch in self.backward_branches[::-1]:
            branch_dict = {
                branch.ID: {
                    "ID": branch.ID,
                    "mother_ID": branch.mother_ID,
                    "points_numsleft_steps": branch.points_numsleft_steps(),
                    "a1a2a3_coefficients": branch.a1a2a3_coefficients,
                    "overshoot": branch.overshoot,
                    "angular_deflection": branch.angular_deflection,
                }
            }
            export_backward_branches = export_backward_branches | branch_dict
        export_backward_branches = { "backward_branches": {**export_backward_branches} }
        
        to_export = export_general | export_trimmer | export_backward_bifurcations | export_backward_branches
        with open(self.exp_name + ".json", "w", encoding="utf-8") as f:
            json.dump(to_export, f, ensure_ascii=False,
                      indent=4, cls=NumpyEncoder)

    def import_branches(self, network):
        """Reorganizing and importing branches from the networks.
        
        Parameters
        ----------
        network : Network

        Returns
        -------
        backward_branches : list
        backward_bifurcations : BackwardBifurcations
        
        """
        network.active_branches = []
        network.sleeping_branches = []
        
        branch_connectivity_0 = network.branch_connectivity.copy()
        seeds_connectivity_0 = network.box.seeds_connectivity.copy()
        
        backward_branches = []
        mother_IDs = []
        for i, forward_branch in enumerate(network.branches):
            network.branch_connectivity[branch_connectivity_0 ==
                                        forward_branch.ID] = i
            network.box.seeds_connectivity[seeds_connectivity_0[:, 1]
                                           == forward_branch.ID, 1] = i
            forward_branch.ID = i
            
            if network.branch_connectivity.size==0:
                mother_ID = -1
                network.active_branches.append(forward_branch)
            else:
                mother = network.branch_connectivity[
                    network.branch_connectivity[:,1]==i, 0]
                if mother.size > 0:
                    mother_ID = mother[0]
                    mother_IDs.append(mother[0])
                else:
                    mother_ID = -1
                
            # forward/backward branch ID is its index in backward_branches
            backward_branches.append(BackwardBranch(ID=i, mother_ID=mother_ID))
            
            # free node ==> active branch
            if network.branch_connectivity.size!=0 and \
                np.isin(forward_branch.ID, network.branch_connectivity[:,0], invert=True):
                network.active_branches.append(forward_branch)

        # above, we go over branches, so the mother_IDs are repeated; hence np.unique
        backward_bifurcations = BackwardBifurcations(np.unique(mother_IDs))

        return backward_branches, backward_bifurcations

    def __compare_networks(self, initial_network, backward_network, test_network, a1a2a3_coefficients):
        """Gathering BEA metrics.
        
        Parameters
        ----------
        initial_network : Network
        backward_network : Network
            Network after backward step.
        test_network : Network
            Network after backward-forward step.
        a1a2a3_coefficients : array
            A 3-n array with a1,a2,a3 coefficients after backward step.

        Returns
        -------
        None.
        
        """

        for i in range(len(initial_network.active_branches)):
            backward_branch = self.backward_branches[initial_network.active_branches[i].ID]
            bifurcation_ind = self.backward_bifurcations.mother_IDs == \
                backward_branch.mother_ID
                
            # if we've reached the bifurcation point for the second time
            if self.backward_bifurcations.flags[bifurcation_ind]==2:
                self.backward_bifurcations.flags[bifurcation_ind] = 3
                self.backward_bifurcations.a1a2a3_coefficients[bifurcation_ind] = a1a2a3_coefficients[i]
                # length_mismatch is saved in the trimmer after first visit in the bifurcation point
        
            # if we haven't reached the bifurcation point
            # (if we've reached it only once then the branched is popped from active_branches)
            else:
                initial_point = initial_network.active_branches[i].points[-1]
                back_point = backward_network.active_branches[i].points[-1]
                test_point = test_network.active_branches[i].points[-1]
    
                # for angular deflection
                v1 = initial_point - back_point
                real_angle = np.arctan2(v1[1], v1[0])
                v2 = test_point - back_point
                test_angle = np.arctan2(v2[1], v2[0])
    
                # overshoot
                backward_branch.overshoot = np.append(backward_branch.overshoot, \
                                                      np.linalg.norm(test_point - initial_point))
                # angular deflection
                backward_branch.angular_deflection = np.append(backward_branch.angular_deflection, \
                                                               real_angle-test_angle)
                # a1, a2, a3 coefficients
                backward_branch.a1a2a3_coefficients = np.vstack((backward_branch.a1a2a3_coefficients, a1a2a3_coefficients[i]))
                
    def run_BEA(self):
        """Run the Backward Evolution Algorithm.

        Returns
        -------
        None.

        """
        start_clock = time.time()
        backward_dts = np.empty(self.back_forth_steps_thresh)
        # while branches list is not empty
        while not len(self.system.network.branches)==1 and self.BEA_step < self.BEA_step_thresh:
            self.BEA_step = self.BEA_step + 1
            print(
                "\n-------------------   Backward Evolution Algorithm step: {step:.0f}   -------------------\n".format(
                    step=self.BEA_step
                )
            )
            print("Date and time: ", datetime.datetime.now())

            ##### BACKWARD STEPS #####
            for i in range(self.back_forth_steps_thresh):
                initial_network, self.system.network, \
                self.backward_branches, self.backward_bifurcations, \
                backward_dts[i] = \
                    self.trimmer.trim(
                        self.system.network, self.backward_branches, \
                        self.backward_bifurcations, self.BEA_step)

            # if there are no living branches we skip forward steps
            if not self.system.network.active_branches:
                continue

            ##### FORWARD STEPS #####
            test_network = self.system.network.copy()
            for i in range(self.back_forth_steps_thresh):
                self.system.extender.ds = backward_dts[i]
                _, a1a2a3_coefficients = \
                    self.system.trajectory_integrator.integrate(
                        extender=self.system.extender,
                        network=test_network,
                        is_BEA_off=False
                )
                test_network.move_tips(step=self.system.growth_gauges[0])

            # compare the network before and after the backward-forward steps
            self.__compare_networks(
                initial_network, self.system.network, test_network, a1a2a3_coefficients)

            if not self.BEA_step % self.dump_every:
                self.export_json()
                # self.system.export_json()

        self.export_json()
        # self.system.export_json()

        print(
            "\n End of the Backward Evolution Algorithm. Time: {clock:.2f}s".format(
                clock=time.time() - start_clock
            )
        )        



"""Classes to integrate tip trajectories.

Classes:
    ModifiedEulerMethod

"""

import numpy as np


class ModifiedEulerMethod:
    """A class to integrate tip trajectories with modified Euler's method.
    
    Modified Euler's method [Ref3]_ (x(n+1) needs to be found implicitly):

    .. math:: x(n+1) = x(n) + dt \cdot 0.5 \cdot (v[x(n)] + v[x(n+1)]).
    
    Attributes
    ----------
    max_approximation_step : int, default 3
        Number of approximation steps:
            - 0:  explicit Euler's method
            - 1:  Heuns's method
            - >1: modified Euler's method    
            
    References
    ----------
    .. [Ref3] https://en.wikipedia.org/wiki/Trapezoidal_rule_(differential_equations)
    
    """
    
    def __init__(self, max_approximation_step=3):
        """Initialize ModifiedEulerMethod.
        
        Parameters
        ----------
        max_approximation_step : int, default 3
        
        Returns
        -------
        None.
        
        """
        self.max_approximation_step = max_approximation_step
        
    def move_test_tips(self, network, dRs):
        """Move test tips (no bifurcations or killing).

        Assign ``dRs`` to each branch in ``network.active_branches`` and extend them.

        """
        for i, branch in enumerate(network.active_branches):
            branch.dR = dRs[i]
            branch.extend()

    def integrate(self, extender, network, is_BEA_off=True):
        """Integrate tip trajectories with modified Euler's method.
    
        Parameters
        ----------
        extender : Extender
            An object of one of the classes from reticuler.extending_kernels.extenders.
        network : Network
            An object of class Network.
        is_BEA_off : bool, default True
            A boolean condition if the Backward Evolution Algorithm is off.
            If True ``a1a2a3_coefficients`` (from the tip position before extension)
            are not returned.
    
        Returns
        -------
        dt : float
            Time of growth at the current evolution step.
        a1a2a3_coefficients_0 : array
            An array of a_i coefficients for each tip in the network.
    
        """
        # x[n + 1] = x[n] + dt * v[x(n)]: finding position n+1 with explicit Euler
        dRs_0 = extender.find_test_dRs(network, is_BEA_off)
        
        # in the BEA we want to move over dt as in the backward step,
        # so no normalization
        a1a2a3_coefficients_0 = extender.pde_solver.a1a2a3_coefficients.copy()
            
        # checking if branches are_moving or each branch.is_bifurcating
        are_moving = extender.check_bifurcation_and_moving_conditions(network)
        dRs_0 = dRs_0[are_moving]
        # print('a1, a2, a3:\n',extender.pde_solver.a1a2a3_coefficients)
    
        # moving test_system by dRs_0
        test_network = network.copy()
        self.move_test_tips(test_network, dRs_0)
    
        dRs_test = 0
        approximation_step = 0
        epsilon_thresh = (
            np.sum(np.linalg.norm(dRs_0, axis=1)) / 1000
        )  # probably to optimize
        epsilon = epsilon_thresh + 1
        # APPROXIMATION LOOP - we end when approximating isn't getting any
        # better (epsilon is the measure) or after 'max_approximation_step' steps
        while epsilon > epsilon_thresh and approximation_step < self.max_approximation_step:
            approximation_step = approximation_step + 1
    
            # v[ x(n+1)] ]: finding velocity at the next point
            dRs_1 = extender.find_test_dRs(test_network, is_BEA_off)
            
            # calculating approximation error and average dR
            epsilon = np.sum(np.linalg.norm(dRs_test - \
                            (dRs_0 + dRs_1) / 2, axis=1)) / len(dRs_0)
            dRs_test = (dRs_0 + dRs_1) / 2
            # print(dRs_test)
            if is_BEA_off:
                dRs_test = dRs_test * extender.ds / np.max(np.linalg.norm(dRs_test, axis=1))
                
            # moving test_system by dR
            test_network = network.copy()
            self.move_test_tips(test_network, dRs_test)
            
            # print('Forth loop, approximation step: {step}, epsilon = {eps}'.format(step=approximation_step, eps=epsilon) )
        if is_BEA_off:
            # normally division dX/a1^eta would give single dt
            # due to the modified Euler's algorithm (dR = (dRs_0+dRs_1)/2 )
            # dt is not perfectly the same for different tips, so we take a mean
            dt = np.mean( np.linalg.norm(dRs_test, axis=1) / extender.pde_solver.a1a2a3_coefficients[...,0]**extender.eta)
        else: 
            dt = extender.ds
        
        # print('dRs: ', dRs_test)
        extender.assign_dRs(network, dRs_test)
        
        return dt, a1a2a3_coefficients_0



"""Preparing plots.

Functions:
    plot_tree(ax, network, height=2.0, width=2.0, \*\*kwargs_tree_plot)

"""
import numpy as np
import matplotlib.pyplot as plt

plt.rcParams.update(
    {
        "figure.dpi": 150,
        "text.usetex": True,
        "xtick.labelsize": 7,
        "ytick.labelsize": 7,
        "axes.linewidth": 0.5,
        "axes.titlesize": 12,
        "axes.labelsize": 8,
        "lines.linewidth": 1,
        "lines.markersize": 1,
        "lines.marker": "None",
        "lines.solid_capstyle": "round",
        "font.family": "serif",
        "font.serif": ["Computer Modern"],  # ['Times New Roman']
    }
)
# 'font.size': 20,
# 'font.family': 'serif',
# 'font.serif': ['Times New Roman']})
colours = plt.rcParams["axes.prop_cycle"].by_key()["color"]
cm2inch = 1 / 2.54
golden = (1 + 5**0.5) / 2


def plot_tree(ax, network, ylim=None, xlim=2.0, **kwargs_tree_plot):
    """Plot a tree with optional arguments from \*\*kwargs_tree, and a black box.

    Parameters
    -------
    ax : Axes
        An object to plot on.
    network : object of class Network
        Network to plot.
    ylim : float, default None
        Plot extends from y=0 to y=``ylim``.
        If None ylim = max( 2, max height of the network ).
    xlim : float, default 2.0
        Plot extends from x=0 to x=``xlim``.
    kwargs_plots : dict, default {`color`: `#0066CC`, `linewidth`: 2.5}
        Arguments to plot the tree.

    Returns
    -------
    None.

    """
    options_tree_plot = {"color": "#0066CC", "linewidth": 2.5}
    options_tree_plot.update(kwargs_tree_plot)

    # PLOT LINES
    y_max = 2
    for branch in network.branches:
        line = branch.points
        y_max = np.max((np.max(line[:, 1]), y_max))
        ax.plot(*line.T, **options_tree_plot)
    # PLOT BOX
    points_to_plot = network.box.points[network.box.connections]
    for pts in points_to_plot:
        ax.plot(*pts.T, linewidth=options_tree_plot["linewidth"] * 2, color="0")

    ax.axis("off")
    # colouring background
    ax.add_artist(ax.patch)
    ax.patch.set_zorder(-1)
    ax.set_facecolor("#def1ff")
    ax.set_aspect("equal")
    ax.set_xlim(0, xlim)
    ylim = 1.05*y_max if ylim is None else ylim
    ax.set_ylim(0, ylim)    



"""Command line script to run the simulation"""

import argparse
import json
import textwrap
import importlib.metadata
import matplotlib.pyplot as plt

from reticuler.system import Box, Network, System
from reticuler.extending_kernels import extenders, pde_solvers, trajectory_integrators
from reticuler.user_interface import graphics

# %%
def main():
    parser = argparse.ArgumentParser(
        description="Grow a network.", formatter_class=argparse.RawTextHelpFormatter
    )

    # defining arguments for parser object
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=importlib.metadata.version("reticuler"),
    )

    parser.add_argument(
        "-in",
        "--input_file",
        type=str,
        nargs=1,
        metavar="exp_name",
        help=textwrap.dedent(
            """\
            File to import. If None, the System is prepared based on the rest of the arguments.
            default = None"""
        ),
        default=None,
    )

    parser.add_argument(
        "-out",
        "--output_file",
        type=str,
        nargs=1,
        metavar="exp_name",
        help=textwrap.dedent(
            """\
            File to export. If we import a file and leave this as default, 
            `system.exp_name` will be set to `input_file`.
            default = '' """
        ),
        default=[""],
    )

    # Growth options
    parser.add_argument(
        "--growth_params",
        type=json.loads,
        nargs=1,
        metavar="dict",
        help=textwrap.dedent(
            """\
            Optional growth parameters.
            
            Pass dictionary in a form (no spaces, 
            backslash before quotes around `value`): 
                "{\"value\":key}"
            default = {} (keeps default values as listed below)
            
            """
        )
        + textwrap.dedent(
            System.__doc__[
                System.__doc__.find("growth_thresh_type")
                - 4 : System.__doc__.find("exp_name")
            ]
        ),
        default=[{}],
    )

    # Box options
    parser.add_argument(
        "-ic",
        "--initial_condition",
        type=int,
        nargs=1,
        metavar="label",
        help=textwrap.dedent(
            Box.construct.__doc__[
                Box.construct.__doc__.find("initial_condition")
                - 8 : Box.construct.__doc__.find("kwargs_construct")
            ]
        ),
        default=[0],
    )
    parser.add_argument(
        "--kwargs_box",
        type=json.loads,
        nargs=1,
        metavar="dict",
        help=textwrap.dedent(
            """\
            Kwargs for Box construct method.
            
            Pass dictionary in a form (no spaces, 
            backslash before quotes around `value`): 
                "{\"value\":key}"
            default = {} (keeps default values as listed below)
            
            """
        )
        + textwrap.dedent(
            Box.construct.__doc__[
                Box.construct.__doc__.find("kwargs_construct")
                - 8 : Box.construct.__doc__.find("Returns")
            ]
        ),
        default=[{}],
    )

    # Trajectory integrator
    parser.add_argument(
        "--trajectory_integrator",
        type=str,
        nargs=1,
        metavar="name",
        help=textwrap.dedent(
            """\
            Trajectory integrator
            default = ModifiedEulerMethod"""
        ),
        default=["ModifiedEulerMethod"],
    )

    # Solver
    parser.add_argument(
        "--pde_solver",
        type=str,
        nargs=1,
        metavar="name",
        help=textwrap.dedent(
            """\
            PDE solver
            default = FreeFEM"""
        ),
        default=["FreeFEM"],
    )
    parser.add_argument(
        "--pde_solver_params",
        type=json.loads,
        nargs=1,
        metavar="dict",
        help=textwrap.dedent(
            """\
            Optional parameters for solver.
            
            Pass dictionary in a form (no spaces, 
            backslash before quotes around `value`): 
                "{\"value\":key}"
            default = {} (keeps default values as listed below)
            
            """
        )
        + "1. FreeFEM\n"
        + textwrap.dedent(
            pde_solvers.FreeFEM.__doc__[
                pde_solvers.FreeFEM.__doc__.find("equation")
                - 4 : pde_solvers.FreeFEM.__doc__.find("a1a2a3_coefficients")
            ]
        ),
        default=[{}],
    )

    # Extender
    parser.add_argument(
        "--extender",
        type=str,
        nargs=1,
        metavar="name",
        help=textwrap.dedent(
            """\
            Extender
            default = Streamline"""
        ),
        default=["Streamline"],
    )
    parser.add_argument(
        "--extender_params",
        type=json.loads,
        nargs=1,
        metavar="dict",
        help=textwrap.dedent(
            """\
            Optional parameters for extender.
            
            Pass dictionary in a form (no spaces, 
            backslash before quotes around `value`): 
                "{\"value\":key}"
            default = {} (keeps default values as listed below)
            
            """
        )
        + "1. Streamline\n"
        + textwrap.dedent(
            extenders.Streamline.__doc__[
                extenders.Streamline.__doc__.find("eta")
                - 4 : extenders.Streamline.__doc__.find("References")
            ]
        ),
        default=[{}],
    )

    # Plotting at the end
    parser.add_argument(
        "-fp",
        "--final_plot",
        action=argparse.BooleanOptionalAction,
        help=textwrap.dedent(
            """\
            Flag indicating to plot the final network.
            """
        ),
    )

    # parse the arguments from standard input
    args = parser.parse_args()

    if args.input_file is None:
        # Prepare the System from scratch

        # Box
        box, branches = Box.construct(
            initial_condition=args.initial_condition[0], **args.kwargs_box[0]
        )

        # Network
        network = Network(box=box, branches=branches, active_branches=branches.copy())

        # Trajectory integrator
        if args.trajectory_integrator[0] == "ModifiedEulerMethod":
            trajectory_integrator = trajectory_integrators.ModifiedEulerMethod()

        # Solver
        if args.pde_solver[0] == "FreeFEM":
            pde_solver = pde_solvers.FreeFEM(**args.pde_solver_params[0])

        # Extender
        if args.extender[0] == "Streamline":
            extender = extenders.Streamline(
                pde_solver=pde_solver, **args.extender_params[0]
            )

        # General
        system = System(
            network=network,
            extender=extender,
            trajectory_integrator=trajectory_integrator,
            exp_name=args.output_file[0],
            **args.growth_params[0]
        )

    else:
        # Import System from JSON file
        system = System.import_json(input_file=args.input_file[0])
        if args.output_file[0] != "":
            system.exp_name = args.output_file[0]

    system.evolve()

    if args.final_plot:
        fig, ax = plt.subplots()
        graphics.plot_tree(ax, network=system.network)
        fig.savefig(system.exp_name + ".jpg", bbox_inches="tight")


if __name__ == "__main__":
    main()



"""Command line script to run the Backward Evolution Algorithm"""

import argparse
import json
import textwrap
import importlib.metadata

from reticuler.system import System
from reticuler.backward_evolution.system_back import BackwardSystem
from reticuler.backward_evolution import trimmers

# %%
def main():
    parser = argparse.ArgumentParser(
        description="Run the Backward Evolution Algorithm.", formatter_class=argparse.RawTextHelpFormatter
    )

    # defining arguments for parser object
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=importlib.metadata.version("reticuler"),
    )
    parser.add_argument(
        "input_file",
        type=str,
        nargs=1,
        metavar="file_name",
        help=textwrap.dedent(
            """\
                            File to import"""
        ),
    )
    parser.add_argument(
        "-out",
        "--output_file",
        type=str,
        nargs=1,
        metavar="exp_name",
        help=textwrap.dedent(
            """\
            File to export:
            exp_name = `exp_name`+'_back.json'.
            If left as default: 
            exp_name = `input_file`+'_back.json'.
            
            default = ''
            """
        ),
        default=[""],
    )

    # BEA options
    parser.add_argument(
        "--BEA_params",
        type=json.loads,
        nargs=1,
        metavar="dict",
        help=textwrap.dedent(
            """\
            Optional BEA parameters.
            
            Pass dictionary in a form (no spaces, 
            backslash before quotes around `value`): 
                "{\"value\":key}"
            default = {} (keeps default values as listed below)
            
            """
        )
        + textwrap.dedent(
            BackwardSystem.__doc__[
                BackwardSystem.__doc__.find("BEA_step_thresh")
                - 4 : BackwardSystem.__doc__.find("exp_name")
            ]
        ),
        default=[{}],
    )

    # Trimmer
    parser.add_argument(
        "--trimmer",
        type=str,
        nargs=1,
        metavar="name",
        help=textwrap.dedent(
            """\
            Trimmer
            default = BackwardModifiedEulerMethod"""
        ),
        default=["BackwardModifiedEulerMethod"],
    )
    parser.add_argument(
        "--trimmer_params",
        type=json.loads,
        nargs=1,
        metavar="dict",
        help=textwrap.dedent(
            """\
            Optional parameters for trimmer.
            
            Pass dictionary in a form (no spaces, 
            backslash before quotes around `value`): 
                "{\"value\":key}"
            default = {} (keeps default values as listed below)
            
            """
        )
        + textwrap.dedent(
            trimmers.BackwardModifiedEulerMethod.__doc__[
               trimmers.BackwardModifiedEulerMethod.__doc__.find("eta")
                - 4 : trimmers.BackwardModifiedEulerMethod.__doc__.find("References")-6
            ]
        ),
        default=[{}],
    )       

    # Continuation
    parser.add_argument(
        "-cont",
        "--continuation_file",
        type=str,
        nargs=1,
        metavar="file_name",
        help=textwrap.dedent(
            """\
            Continuation of the previously commenced BEA.
            System will be imported from `input_file` and backward system from 
            `continuation_file`+'.json'.
            """
        ),
        default=[""],
    )
    
    # parse the arguments from standard input
    args = parser.parse_args()
    
    # Import System from JSON file
    system = System.import_json(input_file=args.input_file[0])
    if args.output_file[0]!="":
        system.exp_name = args.output_file[0]
    else:
        system.exp_name = args.input_file[0]

    # Create or import BackwardSystem
    if args.continuation_file[0]!="":
        backward_system = BackwardSystem.import_json(input_file=args.continuation_file[0], system=system)
        backward_system.exp_name = backward_system.exp_name + "_cont"
    else:
        # Trimmer
        if args.trimmer[0] == "BackwardModifiedEulerMethod":
            trimmer = trimmers.BackwardModifiedEulerMethod(system.extender.pde_solver, **args.trimmer_params[0])
        # All
        backward_system = BackwardSystem(system, trimmer)
        
    # Running BEA
    backward_system.run_BEA()

if __name__ == "__main__":
    main()



"""Clipping a network.

Functions:
    clip_to_step(system, max_step)
    clip_to_length(system, max_length)
    clip_to_height(system, max_height)

"""
import numpy as np

from reticuler.system import System
from reticuler.backward_evolution.system_back import BackwardSystem, BackwardBranch

def clip_to_step(system, max_step):
    branches_to_iterate = system.network.branches.copy()
    for branch in branches_to_iterate:
        to_trash  = branch.steps > max_step
        if sum(to_trash)+1 >= len(branch.steps):
            mask = system.network.branch_connectivity[:,1]!=branch.ID
            if sum(~mask)==1:
                mother_ID = system.network.branch_connectivity[~mask,0][0]
                mother_branch = [b for b in branches_to_iterate if b.ID==mother_ID][0]
                if not mother_branch in system.network.active_branches:
                    system.network.active_branches.append(mother_branch)
            system.network.branch_connectivity = system.network.branch_connectivity[mask,...]
            system.network.branches.remove(branch)
            if branch in system.network.active_branches:
                system.network.active_branches.remove(branch)
            if branch in system.network.sleeping_branches:
                system.network.sleeping_branches.remove(branch)
        else:
            branch.points = branch.points[~to_trash]
            branch.steps = branch.steps[~to_trash]
            
    system.growth_gauges[0] = max_step
    system.growth_gauges[1], system.growth_gauges[2] = system.network.height_and_length()
    system.growth_gauges[3] = 0


def clip_to_length(system, max_length):
    length_throughout_evolution = np.zeros(int(system.growth_gauges[0])+1)
    branches_to_iterate = system.network.branches.copy()
    for branch in branches_to_iterate:
        length_throughout_evolution[branch.steps[1:]] = \
            length_throughout_evolution[branch.steps[1:]] + \
            np.linalg.norm(branch.points[1:]-branch.points[:-1], axis=1)
    length_throughout_evolution = np.cumsum(length_throughout_evolution)
    
    clip_to_step(system, sum(length_throughout_evolution<max_length))

def clip_to_height(system, max_height):
    all_points_steps = np.zeros(3)
    branches_to_iterate = system.network.branches.copy()
    for branch in branches_to_iterate:
        all_points_steps = np.vstack( ( all_points_steps, branch.points_steps() ) )
    all_points_steps = all_points_steps[1:]
    
    clip_to_step(system, np.min(all_points_steps[all_points_steps[:,1]>max_height, 2]))
    
def clip_to_BEA_step(system, backward_system, max_BEA_step):
    branches_to_iterate = backward_system.system.network.branches.copy()
    # active_inds = np.empty(len(branches_to_iterate), dtype=int)
    # sleeping_inds = np.empty(len(branches_to_iterate), dtype=int)
    max_BEA_step_found = 0
    for i, forward_branch in enumerate(branches_to_iterate):
        backward_branch = backward_system.backward_branches[forward_branch.ID]
        # active_inds[i] = backward_branch.active_ind
        # sleeping_inds[i] = backward_branch.sleeping_ind
        if len(backward_branch.steps):
            ind_backward = np.sum(backward_branch.steps<=max_BEA_step)-1
            if ind_backward<0:
                # branch hasn't been trimmed at all -- delete BEA results
                backward_system.backward_branches[forward_branch.ID] =\
                    BackwardBranch(backward_branch.ID, backward_branch.mother_ID)
                # and update backward_bifurcations
                ind_bifurcation = backward_system.backward_bifurcations.mother_IDs==backward_branch.mother_ID
                if backward_system.backward_bifurcations.flags[ind_bifurcation] == 3:
                    backward_system.backward_bifurcations.flags[ind_bifurcation] = 1
                    backward_system.backward_bifurcations.a1a2a3_coefficients[ind_bifurcation,...] = 0
                elif backward_system.backward_bifurcations.flags[ind_bifurcation] < 3:
                    backward_system.backward_bifurcations.flags[ind_bifurcation] = 0
                    backward_system.backward_bifurcations.length_mismatch[ind_bifurcation] = 0
            else:
                max_BEA_step_found = max(max_BEA_step_found, \
                                         min(max_BEA_step, backward_branch.steps[ind_backward]))
                # max_BEA_step_found = 1 
                num_left = backward_branch.nums_left[ind_backward]
                if num_left==0:
                    # whole branch trimmed - delete forward_branch
                    mask = backward_system.system.network.branch_connectivity[:,1]!=forward_branch.ID
                    backward_system.system.network.branch_connectivity = backward_system.system.network.branch_connectivity[mask,...]
                    backward_system.system.network.branches.remove(forward_branch)
                else:
                    # branch partially trimmed - trim accordingly
                    forward_branch.steps = forward_branch.steps[:num_left+1]
                    forward_branch.points = forward_branch.points[:num_left+1]
                    forward_branch.points[-1] = backward_branch.points[ind_backward]
                    backward_branch.points = backward_branch.points[:ind_backward+1]
                    backward_branch.nums_left = backward_branch.nums_left[:ind_backward+1]
                    backward_branch.steps = backward_branch.steps[:ind_backward+1]
                    backward_branch.a1a2a3_coefficients = backward_branch.a1a2a3_coefficients[:ind_backward+1]
                    backward_branch.overshoot = backward_branch.overshoot[:ind_backward+1]
                    backward_branch.angular_deflection = backward_branch.angular_deflection[:ind_backward+1]
                    # and update backward_bifurcations
                    ind_bifurcation = backward_system.backward_bifurcations.mother_IDs==backward_branch.mother_ID
                    if backward_system.backward_bifurcations.flags[ind_bifurcation] == 3:
                        backward_system.backward_bifurcations.flags[ind_bifurcation] = 1
                        backward_system.backward_bifurcations.a1a2a3_coefficients[ind_bifurcation,...] = 0
                    elif backward_system.backward_bifurcations.flags[ind_bifurcation] == 1:
                        backward_system.backward_bifurcations.flags[ind_bifurcation] = 0
                        backward_system.backward_bifurcations.length_mismatch[ind_bifurcation] = 0
                    # print(backward_branch.active_ind)
    backward_system.import_branches(backward_system.system.network)
    backward_system.BEA_step = max_BEA_step_found
    # backward_system.system.network.active_branches = [branches_to_iterate[i] for i in np.argsort(active_inds)[np.sort(active_inds)>-1]]
    # backward_system.system.network.sleeping_branches = [branches_to_iterate[i] for i in np.argsort(sleeping_inds)[np.sort(sleeping_inds)>-1]]
    

if __name__ == "__main__":
    # Import System from JSON file
    dir_name = "G:/My Drive/Research/Network simulations/reticuler/archive/BEA/testing/test_1.0/matlab_comparison/laplace_freefem2_v5/eta_original18/debug_eta0/"
    system = System.import_json(input_file=dir_name+"original_tree_0")
    backward_system = BackwardSystem.import_json(input_file=dir_name+"eta00_back_0", system=system)
    backward_system.exp_name = dir_name+"eta00_back"
    backward_system.system.exp_name = dir_name+"original_tree"

    clip_to_BEA_step(system, backward_system, max_BEA_step=23)
    print(backward_system.backward_branches[-1].steps)

    # backward_system.system.export_json()         
    # backward_system.export_json()      
        
    # import matplotlib.pyplot as plt
    # from reticuler.user_interface import graphics
    # fig, ax = plt.subplots()
    # graphics.plot_tree(
    #     ax,
    #     network=system.network,
    #     ylim=2.5,
    #     # color='{}'.format(step/lim)
    # )    
    
    # Plot different stages of the evolution
    # import matplotlib.pyplot as plt
    # import matplotlib as mpl
    # from reticuler.user_interface import graphics
    # fig, ax = plt.subplots()
    # lim = system.growth_gauges[0]+1
    # for step in reversed(np.linspace(100, lim, 30)):
    #     # print(step)
    #     clip_to_step(system, step)
    #     graphics.plot_tree(
    #         ax,
    #         network=system.network,
    #         ylim=10.5,
    #         color=mpl.colormaps['copper'](step/lim),
    #         # color='{}'.format(step/lim)
    #     )
    


"""Command line script to plot a network"""

import argparse
import json
import textwrap
import matplotlib.pyplot as plt

from reticuler.system import System
from reticuler.user_interface import graphics

# %%
def main():
    parser = argparse.ArgumentParser(
        description="Plot a network.", formatter_class=argparse.RawTextHelpFormatter
    )

    # defining arguments for parser object
    parser.add_argument(
        "input_file",
        type=str,
        nargs=1,
        metavar="file_name",
        help=textwrap.dedent(
            """\
                            File to import"""
        ),
    )
    parser.add_argument(
        "-out",
        "--output_file",
        type=str,
        nargs=1,
        metavar="file_name",
        help=textwrap.dedent(
            """\
                            File to export. If None the same as input.
                            default = None """
        ),
        default=None,
    )

    parser.add_argument(
        "-out_ext",
        "--output_extension",
        type=str,
        nargs=1,
        metavar="ext",
        help=textwrap.dedent(
            """\
                            Output extension ('.pdf', '.svg', '.png', etc.)
                            default = '.jpg' """
        ),
        default=[".jpg"],
    )

    # Plotting options
    parser.add_argument(
        "--ylim",
        type=float,
        nargs=1,
        metavar="num",
        help=textwrap.dedent(
            """\
                            ylim of the plot. If None ylim=max height of the network.
                            default = None
                            """
        ),
        default=[None],
    )
    parser.add_argument(
        "--xlim",
        type=float,
        nargs=1,
        metavar="num",
        help=textwrap.dedent(
            """\
                            xlim of the plot.
                            default = 2
                            """
        ),
        default=[2],
    )
    parser.add_argument(
        "--plot_params",
        type=json.loads,
        nargs=1,
        metavar="dict",
        help=textwrap.dedent(
            """\
                            Optional plotting parameters.
                            
                            Pass dictionary in a form (no spaces, 
                            backslash before quotes around `value`): 
                                "{\"value\":key}"
                            default = {} (keeps default values as listed below)
                            
                            """
        )
        + textwrap.dedent(
            graphics.plot_tree.__doc__[
                graphics.plot_tree.__doc__.find("kwargs_plots")
                - 4 : graphics.plot_tree.__doc__.find("Returns")
            ]
        ),
        default=[{}],
    )

    # parse the arguments from standard input
    args = parser.parse_args()

    # Import System from JSON file
    system = System.import_json(input_file=args.input_file[0])

    fig, ax = plt.subplots()
    graphics.plot_tree(
        ax,
        network=system.network,
        ylim=args.ylim[0],
        xlim=args.xlim[0],
        **args.plot_params[0]
    )

    if args.output_file is None:
        fig.savefig(args.input_file[0] + args.output_extension[0], bbox_inches="tight")
    else:
        fig.savefig(args.output_file[0] + args.output_extension[0], bbox_inches="tight")


if __name__ == "__main__":
    main()



"""Extenders to find the next position of the tip.

Classes:
    Streamline

"""

import numpy as np


class Streamline:
    """Extender based on the streamline algorithm [Ref1]_.

    Attributes
    ----------
    pde_solver : PDESolver
    eta : float, default 1.0
        The growth exponent (v=a1**eta).
        High values increase competition between the branches.
        Low values stabilize the growth.
    ds : float, default 0.01
        A distance over which the fastest branch in the network
        will move in each timestep.
    bifurcation_type : int, default 0
        - 0: no bifurcations
        - 1: a1 bifurcations (velocity criterion)
        - 2: a3/a1 bifurcations (bimodality criterion)
        - 3: random bifurcations
    bifurcation_thresh : float, default 0
        Threshold for the bifurcation criterion.
    bifurcation_angle : float, default 2pi/5
        Angle between the daughter branches after bifurcation.
        Default angle (72 degrees) corresponds to the analytical solution
        for fingers in a diffusive field.
    inflow_thresh : float, default 0.05
        Threshold to put asleep the tips with less than ``inflow_thresh``
        of max flux/velocity.
    distance_from_bif_thresh : float, default 2.1*``ds``
        A minimal distance the tip has to move from the previous bifurcations
        to split again.

    References
    ----------
    .. [Ref1] "Through history to growth dynamics: backward evolution of spatial networks",
            S. Żukowski, P. Morawiecki, H. Seybold, P. Szymczak, Sci Rep 12, 20407 (2022). 
            https://doi.org/10.1038/s41598-022-24656-x

    """

    def __init__(
        self,
        pde_solver,
        eta=1.0,
        ds=0.01,
        bifurcation_type=0,
        bifurcation_thresh=0,
        bifurcation_angle=2 * np.pi / 5,
        inflow_thresh=0.05
    ):
        """Initialize Streamline.

        Parameters
        ----------
        pde_solver : object of class pde_solvers
        eta : float, default 1.0
        ds : float, default 0.01
        bifurcation_type : int, default 0
        bifurcation_thresh : float, default 0
        bifurcation_angle : float, default 2pi/5

        Returns
        -------
        None.

        """
        self.pde_solver = pde_solver

        self.eta = eta
        self.ds = ds

        self.bifurcation_type = bifurcation_type  # no bifurcations, a1, a3/a1, random
        self.bifurcation_thresh = bifurcation_thresh
        if not bifurcation_thresh:
            if self.bifurcation_type == 1:
                self.bifurcation_thresh = 0.8  # a1 bifurcations
            if self.bifurcation_type == 2:
                self.bifurcation_thresh = -0.1  # a3/a1 bifurcations
            if self.bifurcation_type == 3:
                self.bifurcation_thresh = 3 * ds
                # random bifurcations: bif_probability
        self.bifurcation_angle = bifurcation_angle  # 2*np.pi/5

        # less than `inflow_thresh` of max flux/velocity puts branches asleep
        self.inflow_thresh = inflow_thresh
        self.distance_from_bif_thresh = 2.1 * ds

    def __rotation_matrix(self, angle):
        """Construct a matrix to rotate a vector by an ``angle``.

        Parameters
        ----------
        angle : float

        Returns
        -------
        array
            An 2-2 array.

        Examples
        --------
        >>> rot = self.__rotation_matrix(self, angle)
        >>> rotated_vector = np.dot(rot, vector)

        """
        return np.array(
            [[np.cos(angle), np.sin(angle)], [-np.sin(angle), np.cos(angle)]]
        )

    def check_bifurcation_and_moving_conditions(self, network):
        """Check bifurcation and moving conditions."""

        a1 = self.pde_solver.a1a2a3_coefficients[..., 0]
        max_a1 = np.max(a1)
        # checking which branches are_moving
        # (first condition for low eta, second for high)
        are_moving = np.logical_and(a1/max_a1 > self.inflow_thresh,
                                    (a1/max_a1)**self.eta > self.inflow_thresh)

        # shallow copy of active_branches (creates new list instance, but the elements are still the same)
        branches_to_iterate = network.active_branches.copy()
        for i, branch in enumerate(branches_to_iterate):
            a1 = self.pde_solver.a1a2a3_coefficients[i, 0]
            a3 = self.pde_solver.a1a2a3_coefficients[i, 2]
            if (
                self.bifurcation_type
                and branch.length() > self.distance_from_bif_thresh
            ):
                # the second condition above is used to avoid many bifurcations
                # in almost one point which can occur while ds is very small
                if (self.bifurcation_type == 1 and a1 > self.bifurcation_thresh) or (
                    self.bifurcation_type == 2 and a3 / a1 < self.bifurcation_thresh
                ):
                    branch.is_bifurcating = True
                elif self.bifurcation_type == 3:
                    p = self.bifurcation_thresh * (a1 / max_a1) ** self.eta
                    r = np.random.uniform(0, 1)  # uniform distribution [0,1)
                    if p > r:
                        branch.is_bifurcating = True
                        
            if not are_moving[i]:
                network.sleeping_branches.append(branch)
                network.active_branches.remove(branch)
                print("! Branch {ID} is sleeping !".format(ID=branch.ID))

        return are_moving

    def __streamline_extension(self, beta, dr):
        """Calculate a vector over which the tip is shifted.

        Derived from the fact that the finger proceeds along a unique
        streamling going through the tip.

        Parameters
        ----------
        beta : float
            a1/a2 value
        dr : float
            A distance over which the tip is moving.

        Returns
        -------
        dR : array
            An 1-2 array.

        """
        if np.abs(beta) < 1000:
            y = ((beta**2) / 9) * ((27 * dr / (2 * beta**2) + 1) ** (2 / 3) - 1)
        else:
            y = dr - (9*dr**2)/(4*beta**2) + (27*dr**3) / \
                (2*beta**4) - (1701*dr**4)/(16*beta**6)
        x = np.around(
            np.sign(beta) * 2 * ((y**3 / beta**2) +
                                 (y / beta) ** 4) ** (1 / 2), 9)                                                      
        return np.array([x, y])

    def find_test_dRs(self, network, is_BEA_off):
        """Find a single test shift over which the tip is moving.

        Parameters
        ----------
        network : object of class Network

        Returns
        -------
        dRs_test : array
            An n-2 array with dx and dy shifts for each tip.

        """
        # running PDE solver
        # self.pde_solver.a1a2a3_coefficients are updated in FreeFEM solver
        self.pde_solver.solve_PDE(network)

        if is_BEA_off:
            # normalize dr, so that the fastest tip moves over ds
            dr_norm = np.max(self.pde_solver.a1a2a3_coefficients[..., 0] ** self.eta)
        else:
            # dr_norm = 1, ds <=> dt
            dr_norm = 1
        dRs_test = np.empty((len(network.active_branches), 2))
        for i, branch in enumerate(network.active_branches):
            a1 = self.pde_solver.a1a2a3_coefficients[i, 0]
            a2 = self.pde_solver.a1a2a3_coefficients[i, 1]
            beta = a1 / a2

            # __streamline_extension formula is derived in the coordinate
            # system where the tip segment lies on a negative Y axis;
            # hence, we rotate obtained dR vector to that system
            tip_angle = np.pi / 2 - branch.tip_angle()
            dr = self.ds * a1**self.eta / dr_norm
            dRs_test[i] = np.dot(
                self.__rotation_matrix(
                    tip_angle), self.__streamline_extension(beta, dr)
            )
        return dRs_test

    def assign_dRs(self, network, dRs):
        """Assign ``dRs`` to each branch in ``network``."""
        for i, branch in enumerate(network.active_branches):
            if branch.is_bifurcating:
                branch.dR = [
                    np.dot(
                        self.__rotation_matrix(-self.bifurcation_angle / 2), dRs[i]),
                    np.dot(self.__rotation_matrix(
                        self.bifurcation_angle / 2), dRs[i]),
                ]
            else:
                branch.dR = dRs[i]


# class Royal:
#     def __init__(self, solver)

# class BEM:
#     def __init__(self, geometry):

# class BARW:
#     def __init__(self, geometry):
